"""
Simultaneous model simulation inlay
"""


#[
from __future__ import annotations

import numpy as _np

from typing import Any, Literal
from .. import quantities as _quantities
from .. import has_variants as _has_variants
from .. import wrongdoings as _wrongdoings
from ..import dates as _dates
from ..databoxes.main import Databox
from ..dataslates.main import Dataslate
from ..fords import solutions as _solutions
from ..plans.simulation_plans import SimulationPlan
from ..fords import simulators as _ford_simulator
from ..period_by_period import simulators as _period_by_period_simulator
from ..stacked_time import simulators as _stacked_time_simulator

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from collections.abc import Iterable
    from ..frames import Frame
    from ..dates import Period
#]


_SIMULATOR_MODULE = {
    "first_order": _ford_simulator,
    "period_by_period": _period_by_period_simulator,
    "period": _period_by_period_simulator,
    "stacked_time": _stacked_time_simulator,
    "stacked": _stacked_time_simulator,
}


_Info = dict[str, Any] | list[dict[str, Any]]


def mixin(klass: type, ) -> type:
    r"""
    Mix the simulate method into the class
    """
    #[
    klass.simulate = simulate
    return klass
    #]


#-------------------------------------------------------------------------------
# Functions to be used as methods in Simultaneous class
#-------------------------------------------------------------------------------


def simulate(
    self,
    in_db: Databox,
    span: Iterable[Period],
    *,
    plan: SimulationPlan | None = None,
    method: Literal["first_order", "period_by_period", "period", "stacked_time", "stacked", ] = "first_order",
    prepend_input: bool = True,
    target_db: Databox | None = None,
    num_variants: int | None = None,
    remove_initial: bool = True,
    remove_terminal: bool = True,
    shocks_from_data: bool = True,
    stds_from_data: bool = True,
    parameters_from_data: bool = False,
    output_parameters: bool = False,
    force_split_frames: bool = False,
    when_fails: Literal["critical", "error", "warning", "silent"] = "critical",
    #
    unpack_singleton: bool = True,
    return_info: bool = False,
    **kwargs,
) -> Databox | tuple[Databox, _Info]:
    r"""
    """
    #[
    num_variants = self.resolve_num_variants_in_context(num_variants, )

    base_dates = tuple(span, )

    if plan is not None and not plan.is_empty:
        plan.check_consistency(self, base_dates, )
        extra_databox_names = plan.get_databox_names()
    else:
        plan = None
        extra_databox_names = None

    slatable = self.slatable_for_simulate(
        shocks_from_data=shocks_from_data,
        stds_from_data=stds_from_data,
        parameters_from_data=parameters_from_data,
        output_parameters=output_parameters,
    )

    dataslate = Dataslate.from_databox_for_slatable(
        slatable, in_db, base_dates,
        num_variants=num_variants,
        extra_databox_names=extra_databox_names,
    )

    zipped = zip(
        range(num_variants, ),
        self.iter_variants(),
        dataslate.iter_variants(),
    )

    when_fails_stream = _wrongdoings.create_stream(when_fails, "Simulation failed to complete", )
    unanticipated_shock_qids = _get_unanticipated_shock_qids(self, )
    simulator_module = _SIMULATOR_MODULE[method]

    #=======================================================================
    # Main loop over variants and over frames
    #
    out_info = []
    for vid, model_v, dataslate_v in zipped:

        input_data_array = dataslate_v.get_data_variant().copy()

        frames = simulator_module.create_frames(
            model_v, dataslate_v, plan,
            force_split_frames=force_split_frames,
        )

        info_v = {
            "method": simulator_module.METHOD_NAME,
            "frames": tuple(frames),
            "exit_status": (),
            "frame_databoxes": (),
        }

        simulator_module.simulate_initial_guess(
            model_v, dataslate_v, plan,
            **kwargs,
        )

        for frame in frames:

            simulation_header = _create_simulation_header(vid, frame, )

            # Create a full copy of the dataslate, both invariant and
            # variants. This is necessary because the frame will first prune
            # the data in place, and then remove initial and terminal data
            # and periods.
            frame_ds = dataslate_v.copy()

            # Remove unanticipated data from the frame except the first
            # simulation period for SplitFrames; keep everything unchanged
            # for SingleFrames
            frame.prune_frame_data(frame_ds, unanticipated_shock_qids, )

            exit_status = simulator_module.simulate_frame(
                model_v, frame_ds,
                frame=frame,
                input_data_array=input_data_array,
                plan=plan,
                simulation_header=simulation_header,
                return_info=return_info,
                **kwargs,
            )

            if not exit_status.is_success:
                when_fails_stream.add(f"{simulation_header}: {exit_status}", )

            # In SplitFrames, copy frame data back to the main dataslate
            # * Unanticipated shocks are written back within the frame slice
            # * Other data are writen back within the simulation slice
            # In SingleFrames, do nothing because everyrhing is kept and
            # performed within the main dataslate
            frame.write_frame_data_to_main_dataslate(
                dataslate_v, frame_ds, unanticipated_shock_qids,
            )

            # Create a frame databox only if output info is requested
            # for performance reasons
            frame_db = (
                frame_ds.to_databox(span="base", )
                if return_info else None
            )

            info_v["frame_databoxes"] += (frame_db, )
            info_v["exit_status"] += (exit_status, )

        out_info.append(info_v, )

    when_fails_stream._raise()
    #
    #=======================================================================
    #
    #
    # Remove initial and terminal condition data (all lags and leads
    # before and after the simulation span)
    if remove_terminal:
        dataslate.remove_terminal()
    if remove_initial:
        dataslate.remove_initial()
    #
    # Convert all variants of the dataslate to a databox
    out_db = dataslate.to_databox()
    if prepend_input:
        out_db.prepend(in_db, base_dates[0]-1, )
    #
    # Add to custom databox
    if target_db is not None:
        out_db = target_db | out_db
    #
    if return_info:
        is_singleton = num_variants == 1
        out_info = _has_variants.unpack_singleton(
            out_info, is_singleton,
            unpack_singleton=unpack_singleton,
        )
        return out_db, out_info
    else:
        return out_db
    #]


#-------------------------------------------------------------------------------


def _append_frame_databox(
    frame_databoxes: list[Databox],
    frame_ds: Dataslate,
) -> None:
    """
    """
    #[
    frame_db = frame_ds.to_databox(span="base", )
    frame_databoxes += (frame_db, )
    #]


def _get_unanticipated_shock_qids(self, /, ) -> tuple[int, ...]:
    """
    """
    #[
    plannable = self.get_simulation_plannable()
    name_to_qid = self.create_name_to_qid()
    unanticipated_shock_names = getattr(plannable, "can_be_endogenized_unanticipated", (), )
    return tuple(name_to_qid[i] for i in unanticipated_shock_names)
    #]


def _create_simulation_header(
    vid: str,
    frame: Frame,
) -> str:
    """
    """
    #[
    simulation_span = _dates.get_printable_span(frame.start, frame.simulation_end, )
    return f"[Variant {vid}][Periods {simulation_span}]"
    #]




"""
"""


#[

from __future__ import annotations

import numpy as _np
import scipy as _sp
import itertools as _it
import documark as _dm

from ..incidences import main as _incidences
from ..fords import covariances as _covariances
from .. import quantities as _quantities
from ..namings import DimensionNames

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from numbers import Real
    from collections.abc import Iterable

#]


class Inlay:
    r"""
    """
    #[

    @_dm.reference(category="parameters", )
    def rescale_stds(
        self,
        factor: Real,
        *,
        kind: _quantities.QuantityKind | None = None,
    ) -> None:
        r"""
................................................................................

==Rescale the standard deviations of model shocks==

Adjust the standard deviations of the model shocks by a specified factor. 
This method allows scaling the standard deviations for the shocks in the 
model based on the provided factor.

    self.rescale_stds(
        factor,
        kind=None,
    )


### Input arguments ###


???+ input "factor"
    A real non-negative number by which to scale the standard deviations of the
    model shocks. This value is used to multiply the existing standard
    deviations.

???+ input "kind"
    An optional parameter to narrow down the types of shocks to rescale. It
    can be one, or a combination, of the following:

    * `ir.TRANSITION_STD`
    * `ir.MEASUREMENT_STD`
    * `ir.ANY_STD`

    If `None`, the standard deviations of all shocks will be rescaled.


### Returns ###


???+ returns "None"
    This method does not return any value but modifies the standard deviations 
    of model shocks in-place, rescaling them.

................................................................................
        """
        if 1.0 * factor <= 0:
            raise ValueError("The scaling factor must be a real non-negative number")
        std_qids = self._get_std_qids(kind=kind, )
        for v in self._variants:
            v.rescale_values("levels", factor, std_qids, )

    def get_acov_dimension_names(self, ) -> DimensionNames:
        r"""
        """
        system_vector, *_ = _get_system_vector(self, )
        qid_to_name = self.create_qid_to_name()
        qid_to_logly = self.create_qid_to_logly()
        names = tuple(
            _quantities.wrap_logly(qid_to_name[tok.qid], qid_to_logly[tok.qid], )
            for tok in system_vector
        )
        return DimensionNames(rows=names, columns=names, )

    def get_acov(
        self,
        up_to_order: int = 0,
        unpack_singleton: bool = True,
    ) -> tuple[_np.ndarray, ...] | list[tuple[_np.ndarray, ...]]:
        r"""
        Asymptotic autocovariance of model variables
        """
        #
        # Combine vectors of transition and measurement tokens, and select
        # those with zero shift only using a boolex
        _, boolex_zero_shift, = _get_system_vector(self, )
        #
        # Tuple element cov_by_variant[variant_index][order] is an N-by-N
        # covariance matrix
        cov_by_variant = [
            self.getv_autocov(v, boolex_zero_shift, up_to_order=up_to_order, )
            for v in self._variants
        ]
        #
        # Tuple element cov_by_order[order] is an N-by-N covariance matrix
        # (for singleton models) or an N-by-N-by-num_variants covariance
        # matrix (for nonsingleton models)
        # stack_func = _STACK_FUNC_FACTORY[self.is_singleton]
        # cov_by_order = tuple(
        #     stack_func(tuple(cov[order] for cov in cov_by_variant))
        #     for order in range(up_to_order + 1)
        # )
        #
        cov_by_variant = self.unpack_singleton(
            cov_by_variant,
            unpack_singleton=unpack_singleton,
        )
        #
        return cov_by_variant

    def get_acorr(self, *args, **kwargs, ):
        r"""
        """
        return _covariances.get_acorr_by_variant(self, *args, **kwargs, )

    def getv_autocov(
        self,
        variant: _variants.Variant,
        boolex_zero_shift: tuple[bool, ...] | Ellipsis,
        #
        up_to_order: int = 0,
    ) -> _np.ndarray:
        r"""
        """
        def select(cov: _np.ndarray, ) -> _np.ndarray:
            """
            Select elements of solution vectors with zero shift only
            """
            cov = cov[boolex_zero_shift, :]
            cov = cov[:, boolex_zero_shift]
            return cov
        cov_u = self.getv_cov_u(variant, )
        cov_w = self.getv_cov_w(variant, )
        cov_by_order = _covariances.get_autocov_square(variant.solution, cov_u, cov_w, up_to_order, )
        return tuple(select(cov, ) for cov in cov_by_order)

    def get_stdvec_transition_shocks(self, ):
        r"""
        """
        std_u = [ std_u for std_u in self.iter_std_u() ]
        return self.unpack_singleton(std_u, )

    def iter_std_u(self, ) -> Iterable[_np.ndarray, ...]:
        r"""
        """
        for v in self._variants:
            yield self.getv_std_u(v, )

    def iter_cov_u(self, ) -> Iterable[_np.ndarray, ...]:
        r"""
        """
        for v in self._variants:
            yield self.getv_cov_u(v, )

    def get_stdvec_measurement_shocks(self, ):
        r"""
        """
        std_w = [ std_w for std_w in self.iter_std_w() ]
        return self.unpack_singleton(std_w, )

    def iter_std_w(self, ) -> Iterable[_np.ndarray, ...]:
        r"""
        """
        for v in self._variants:
            yield self.getv_std_w(v, )

    def iter_std_w(self, ) -> Iterable[_np.ndarray, ...]:
        r"""
        """
        for v in self._variants:
            yield self.getv_std_w(v, )

    # TODO: Refactor this function and getv_cov_u to avoid code duplication
    def _gets_cov_transition_shocks(self, ) -> _np.ndarray:
        r"""
        """
        return self.getv_cov_u(self._variants[0], )

    def get_cov_transition_shocks(
        self,
        unpack_singleton: bool = True,
    ):
        r"""
        """
        cov_u = [
            self.getv_cov_u(v, )
            for v in self._variants
        ]
        return self.unpack_singleton(cov_u, unpack_singleton=unpack_singleton, )

    def get_cov_measurement_shocks(
        self,
        unpack_singleton: bool = True,
    ):
        r"""
        """
        cov_w = [
            self.getv_cov_w(v, )
            for v in self._variants
        ]
        return self.unpack_singleton(cov_w, unpack_singleton=unpack_singleton, )

    def getv_std_u(self, variant, ):
        r"""
        """
        shocks = self._invariant.dynamic_descriptor.solution_vectors.transition_shocks
        return _retrieve_stds(self, variant, shocks, )

    def getv_std_w(self, variant, ):
        r"""
        """
        shocks = self._invariant.dynamic_descriptor.solution_vectors.measurement_shocks
        return _retrieve_stds(self, variant, shocks, )

    def getv_cov_u(self, variant, ):
        r"""
        """
        stds = self.getv_std_u(variant, )
        return _np.diag(stds**2, )

    def getv_cov_w(self, variant, ):
        r"""
        """
        stds = self.getv_std_w(variant, )
        return _np.diag(stds**2, )

    #]


def _stack_singleton(x: tuple[_np.ndarray], ) -> _np.ndarray:
    return x[0]


def _stack_nonsingleton(x: tuple[_np.ndarray], ) -> _np.ndarray:
    return _np.dstack(x)


_STACK_FUNC_FACTORY = {
    True: _stack_singleton,
    False: _stack_nonsingleton,
}


def _get_dimension_names(
    self,
    system_vector: tuple[_incidences.Token, ...],
) -> DimensionNames:
    r"""
    """


def _get_system_vector(self, ) -> tuple[tuple[_incidences.Token, ...], tuple[bool, ...]]:
    r"""
    """
    system_vector = \
        self._invariant.dynamic_descriptor.solution_vectors.transition_variables \
        + self._invariant.dynamic_descriptor.solution_vectors.measurement_variables
    boolex_zero_shift = tuple(tok.shift == 0 for tok in system_vector)
    system_vector = tuple(_it.compress(system_vector, boolex_zero_shift, ))
    return system_vector, boolex_zero_shift,


def _retrieve_stds(
    self,
    variant: Variant,
    shock_tokens: Iterable[Token],
) -> _np.ndarray:
    r"""
    """
    #[
    std_qids = tuple(
        self._invariant.shock_qid_to_std_qid[i.qid]
        for i in shock_tokens
    )
    return variant.retrieve_values_as_array("levels", std_qids, )
    #]




"""
"""


#[

from __future__ import annotations

from collections import namedtuple
import functools as _ft
import itertools as _it
import numpy as _np
import scipy as _sp
import neqs as _ne
import warnings as _wa

from .. import equations as _equations
from ..databoxes import Databox
from ..incidences import blazer as _blazer
from ..incidences.blazer import Block, HumanBlock
from .. import dates as _times
from .. import has_variants as _has_variants
from .. import wrongdoings as _wrongdoings
from ..fords import steadiers as _fs
from ..steadiers import evaluators as _evaluators
from ..steadiers import solver_dispatcher as _solver_dispatcher
from ..equations import ENDOGENOUS_EQUATION
from ..quantities import QuantityKind

from . import _flags

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ..plans.steady_plans import SteadyPlan
    from ..equations import Equation
    from ._variants import Variant
    from collections.abc import Iterable, Callable
    from typing import Any, Literal, NoReturn
    from ..steadiers.solver_dispatcher import SolverType

#]


_Wrt = namedtuple("Wrt", (
    "equations",
    "eids",
    "names",
    "fixed_level_names",
    "fixed_change_names",
    "qids",
    "fixed_level_qids",
    "fixed_change_qids",
))


_STEADY_EQUATION_SOLVED = ENDOGENOUS_EQUATION

_DEFAULT_SOLVER = "neqs_levenberg"


def mixin(klass: type, ) -> type:
    r"""
    Inlay the steady state methods in the class
    """
    #[
    klass.solve_steady = solve_steady
    klass.check_steady = check_steady
    klass.split_into_blocks = split_into_blocks
    klass.build_steady_paths = build_steady_paths
    klass.build_zero_paths = build_zero_paths
    #
    # Legacy aliases
    klass.steady = solve_steady
    #
    return klass
    #]


#-------------------------------------------------------------------------------
# Functions to be used as methods in Simultaneous class
#-------------------------------------------------------------------------------


def solve_steady(
    self,
    unpack_singleton: bool = True,
    update_steady_autovalues: bool = True,
    return_info: bool = False,
    **kwargs,
) -> dict | list[dict]:
    r"""
    Calculate steady state for each Variant within this model
    """
    #[
    model_flags = self.resolve_flags(**kwargs, )
    steady_solver = _choose_steady_solver(model_flags.is_linear, model_flags.is_flat, )
    out_info = []
    for vid, v in enumerate(self._variants, ):
        out_info_v = steady_solver(self, v, model_flags, vid, **kwargs, )
        out_info.append(out_info_v, )
    if update_steady_autovalues:
        self.update_steady_autovalues()
    if return_info:
        return _has_variants.unpack_singleton(
            out_info, self.is_singleton,
            unpack_singleton=unpack_singleton,
        )
    #]


def check_steady(
    self,
    equation_switch: Literal["steady", "dynamic"] = "dynamic",
    when_fails: _wrongdoings.HOW = "error",
    tolerance: float | None = None,
    return_info: bool = False,
    unpack_singleton: bool = True,
) -> NoReturn | tuple[bool, tuple[dict, ...]]:
    r"""
    Verify currently assigned steady state in dynamic or steady equations for each variant within this model
    """
    #[
    tolerance = tolerance or self.get_tolerance("equality", )
    qid_to_logly = self.create_qid_to_logly()
    equations = getattr(self._invariant, f"{equation_switch}_equations", )
    equator = self._choose_plain_equator(equation_switch, )
    steady_arrays = (
        v.create_steady_array(
            qid_to_logly,
            num_columns=equator.min_num_columns + 1,
            shift_in_first_column=equator.min_shift,
        ) for v in self._variants
    )
    #
    # REFACTOR
    #
    t_zero = -equator.min_shift
    discrepancies = []
    for x in steady_arrays:
        at_t_zero = equator.eval(x, t_zero, )
        at_t_zero_plus_one = equator.eval(x, t_zero + 1, )
        discrepancies.append(_np.hstack((
            _np.array(at_t_zero).reshape(-1, 1),
            _np.array(at_t_zero_plus_one).reshape(-1, 1),
        )))
    #
    # REFACTOR
    #
    fail_stream = _wrongdoings.create_stream(
        when_fails,
        "Steady state discrepancies in these equations",
    )
    status = []
    failed_discrepancies = []
    failed_equations = []
    for vid, d, in enumerate(discrepancies, ):
        where = _np.any(_np.abs(d) > tolerance, axis=1, ).nonzero()[0].tolist()
        for j in where:
            fail_stream.add(f"[Variant {vid}] {equations[j].human}: {d[j, :].tolist()}")
        status.append(not where)
        failed_discrepancies.append(d[where, :], )
        failed_equations.append(tuple(equations[j].human for j in where), )
    overall_status = all(status)
    #
    fail_stream._raise()
    #
    if not return_info:
        return overall_status
    #
    info = [
        {"discrepancies": d, "failed_discrepancies": f, "failed_equations": e, }
        for d, f, e in zip(discrepancies, failed_discrepancies, failed_equations, )
    ]
    info = _has_variants.unpack_singleton(
        info, self.is_singleton,
        unpack_singleton=unpack_singleton,
    )
    return overall_status, info,
    #]


def split_into_blocks(
    self,
    plan: SteadyPlan,
    **kwargs,
) -> tuple[HumanBlock, ...]:
    r"""
    """
    #[
    model_flags = self.resolve_flags(**kwargs, )
    wrt = _resolve_steady_wrt(self, plan, is_flat=model_flags.is_flat, )
    im = _calculate_steady_incidence_matrix(wrt.equations, wrt.qids, )
    blocks = _blazer.blaze(im, wrt.eids, wrt.qids, )
    equations = self._invariant.steady_equations
    quantities = self._invariant.quantities
    #
    return tuple(
        HumanBlock(i, equations, quantities, )
        for i in blocks
    )
    #]


def build_steady_paths(
    self,
    span: Iterable[Period],
    *,
    deviation: bool = False,
    prepend_initial: bool = True,
    append_terminal: bool = True,
    target_db: Databox | None = None,
    **kwargs,
) -> Self:
    r"""
................................................................................


==Create a Databox with steady-state paths==


Create a Databox with steady-state paths for all model variables, based on the provided
model object and the time span. This method generates steady-state time series
data for each item in the model. This constructor can be used for models that
have well-defined steady state, i.e. Simultaneous models and
VectorAutoregression models.


steady_cb = self.build_steady_paths(
    model,
    span,
    *,
    deviation=False,
    prepend_initial=True,
    append_terminal=True,
    target_db=None,
)


### Input arguments ###


???+ input "model"
The model object for which to generate steady-state time series data.


???+ input "span"
The time span for which to generate steady-state time series data.

???+ input "deviation"
If `True`, the steady-state values are generated as deviations from the
steady state in the form depending on the log status of each variable. If
`False`, the steady-state values are generated in their original level form.

???+ input "prepend_initial"
If `True`, the generated time series will include initial periods before the
start of the specified `span`, to accommodate any lags in the model.

???+ input "append_terminal"
If `True`, the generated time series will include terminal periods after the
end of the specified `span`, to accommodate any leads in the model.

???+ input "target_db"
An optional existing Databox to which the generated steady-state time series
will be added. If `None`, a new Databox will be created.


### Returns ###

???+ returns "steady_databox"
A Databox containing steady-state time series for the `model`.


................................................................................
    """
    start, end = _times.extend_span(
        span,
        self.max_lag,
        self.max_lead,
        prepend_initial,
        append_terminal,
    )
    items = self.generate_steady_items(
        start, end,
        deviation=deviation,
        **kwargs,
    )
    target_db = target_db or Databox()
    target_db.update({ k: v for k, v in items })
    return target_db


build_zero_paths = _ft.partial(
    build_steady_paths,
    deviation=True,
)


#-------------------------------------------------------------------------------


def _steady_linear(
    self,
    variant: Variant,
    model_flags: _flags.Flags,
    vid: int,
    *,
    algorithm: Callable,
    **kwargs,
) -> dict[str, Any]:
    r"""
    """
    #[
    # Calculate first-order system for steady equations for this variant
    system = self._systemize(
        variant,
        self._invariant.steady_descriptor,
        model_flags,
    )
    #
    #=======================================================================
    # Core algorithm: Calculate steady state for this variant
    Xi, Y, dXi, dY = algorithm(system)
    levels = _np.hstack(( Xi.flat, Y.flat )).flatten()
    changes = _np.hstack(( dXi.flat, dY.flat )).flatten()
    #=======================================================================
    #
    # From first-order state space, extract only tokens with zero shift
    tokens = tuple(_it.chain(
        self._invariant.steady_descriptor.system_vectors.transition_variables,
        self._invariant.steady_descriptor.system_vectors.measurement_variables,
    ))
    #
    # True for tokens with zero shift, e.g. [True, False, True, ... ]
    zero_shift_index = [ not t.shift for t in tokens ]
    #
    # Extract steady levels for quantities with zero shift
    levels = levels[zero_shift_index]
    changes = changes[zero_shift_index]
    qids = tuple(t.qid for t in _it.compress(tokens, zero_shift_index))
    #
    # Delogarithmize when needed
    self.delogarithmize(levels, changes, )
    #
    variant.update_levels_from_array(levels, qids, )
    variant.update_changes_from_array(changes, qids, )
    #
    info = {}
    return info
    #]


def _steady_nonlinear(
    self,
    variant: Variant,
    model_flags: _flags.Flags,
    vid: int,
    evaluator_class: type,
    *,
    plan: SteadyPlan | None = None,
    split_into_blocks: bool | None = None,
    solver: SolverType = _DEFAULT_SOLVER,
    solver_settings: dict[str, Any] | None = None,
    iter_printer_settings: dict[str, Any] | None = None,
    **kwargs,
) -> dict[str, Any]:
    r"""
    """
    #[
    qid_to_kind = self.create_qid_to_kind()
    qid_to_name = self.create_qid_to_name()
    #
    split_into_blocks = _resolve_split_into_blocks(split_into_blocks, plan, )
    wrt = _resolve_steady_wrt(self, plan, is_flat=model_flags.is_flat, )
    if split_into_blocks:
        im = _calculate_steady_incidence_matrix(wrt.equations, wrt.qids, )
        blocks = _blazer.blaze(im, wrt.eids, wrt.qids, )
    else:
        blocks = (_blazer.Block(wrt.eids, wrt.qids), )
    num_blocks = len(blocks)
    #
    all_quantities = self.get_quantities()
    all_equations = self.get_steady_equation_objects(kind=_STEADY_EQUATION_SOLVED, )
    human_blocks = _blazer.human_blocks_from_blocks(
        blocks,
        all_equations,
        all_quantities,
    )
    #
    # Prepare the solver and iter printer settings
    #
    solve = getattr(_solver_dispatcher, solver)
    create_solver_settings = getattr(_solver_dispatcher, "create_solver_settings_for_" + solver, )
    solver_settings = create_solver_settings(
        user_solver_settings=solver_settings,
        self_equality_tolerance=self.get_tolerance("equality", ),
    )
    iter_printer_settings = iter_printer_settings or {}
    iter_printer_settings["norm_order"] = solver_settings["norm_order"]
    #
    #
    info_v = {
        "success": None,
        "blocks": [],
    }
    #=======================================================================
    for bid, (block, human_block) in enumerate(zip(blocks, human_blocks, ), ):
        #
        info_b = {
            "success": None,
            "exit_status": None,
            "equations": human_block.equations,
            "quantities": human_block.quantities,
        }
        custom_header = f"[Variant {vid}][Block {bid}]"
        #
        block_level_qids = tuple(sorted(set(block.qids) - set(wrt.fixed_level_qids)))
        block_change_qids = tuple(sorted(set(block.qids) - set(wrt.fixed_change_qids)))
        block_equations = tuple(wrt.equations[eid] for eid in block.eids)
        #
        has_no_qids = (not block_level_qids) and (not block_change_qids)
        has_no_equations = not block_equations
        #
        if has_no_qids or has_no_equations:
            info_b["success"] = True
            info_b["exit_status"] = _ne.ExitStatus.NO_SOLVER_NEEDED
            info_v["blocks"].append(info_b, )
            print(f"\n-Skipping {custom_header}")
            continue
        #
        steady_evaluator = evaluator_class(
            block_level_qids,
            block_change_qids,
            block_equations,
            all_quantities,
            variant,
            context=self._invariant._context,
            iter_printer_settings=iter_printer_settings | {"custom_header": custom_header},
        )
        #
        maybelog_init_guess = steady_evaluator.get_init_guess()
        maybelog_final_guess, success, exit_status = solve(
            steady_evaluator,
            maybelog_init_guess,
            solver_settings=solver_settings,
        )
        steady_evaluator.final_guess = maybelog_final_guess
        #
        if not success:
            _throw_block_error(human_block, custom_header, )
        #
        # Update variant with steady levels and changes
        _update_variant_with_final_guess(variant, steady_evaluator, qid_to_kind, qid_to_name, )
        #
        info_b["success"] = success
        info_b["exit_status"] = exit_status
        info_v["blocks"].append(info_b, )
    #=======================================================================
    #
    #
    info_v["success"] = all(i["success"] for i in info_v["blocks"])
    return info_v
    #]


def _resolve_steady_wrt(
    self,
    plan: SteadyPlan,
    is_flat: bool,
) -> _Wrt:
    r"""
    """
    #[
    wrt_equations = self.get_steady_equation_objects(kind=_STEADY_EQUATION_SOLVED, )
    wrt_eids = tuple(e.id for e in wrt_equations)
    #
    plannable = self.get_steady_plannable(is_flat=is_flat, )
    wrt_names = set(plannable.can_be_exogenized)
    if plan is None or plan.is_empty:
        exogenized_names = ()
        endogenized_names = ()
        fixed_level_names = ()
        fixed_change_names = ()
    else:
        exogenized_names = plan.get_exogenized_names()
        endogenized_names = plan.get_endogenized_names()
        fixed_level_names = plan.get_fixed_level_names()
        fixed_change_names = plan.get_fixed_change_names() + endogenized_names
    #
    wrt_names = (wrt_names - set(exogenized_names)) | set(endogenized_names)
    # wrt_level_names = (wrt_names - set(fixed_level_names))
    # wrt_change_names = (wrt_names - set(fixed_change_names)) if not is_flat else ()
    # wrt_any_names = wrt_level_names | wrt_change_names
    #
    name_to_qid = self.create_name_to_qid()
    # wrt_level_qids = tuple(sorted([name_to_qid[name] for name in wrt_level_names]))
    # wrt_change_qids = tuple(sorted([name_to_qid[name] for name in wrt_change_names]))
    def get_sorted_qids(names: Iterable[str], ) -> tuple[int, ...]:
        return tuple(sorted([name_to_qid[name] for name in names]))
    wrt_qids = get_sorted_qids(wrt_names, )
    wrt_fixed_level_qids = get_sorted_qids(fixed_level_names, )
    wrt_fixed_change_qids = get_sorted_qids(fixed_change_names, )
    #
    # Make order of names consistent with qids
    qid_to_name = self.create_qid_to_name()
    # wrt_level_names = tuple(qid_to_name[qid] for qid in wrt_level_qids)
    # wrt_change_names = tuple(qid_to_name[qid] for qid in wrt_change_qids)
    wrt_names = tuple(qid_to_name[qid] for qid in wrt_qids)
    wrt_fixed_level_names = tuple(qid_to_name[qid] for qid in wrt_fixed_level_qids)
    wrt_fixed_change_names = tuple(qid_to_name[qid] for qid in wrt_fixed_change_qids)
    #
    return _Wrt(
        equations=wrt_equations,
        eids=wrt_eids,
        names=wrt_names,
        fixed_level_names=wrt_fixed_level_names,
        fixed_change_names=wrt_fixed_change_names,
        qids=wrt_qids,
        fixed_level_qids=wrt_fixed_level_qids,
        fixed_change_qids=wrt_fixed_change_qids,
    )
    #]


def _choose_steady_solver(
    is_linear: bool,
    is_flat: bool,
) -> Callable:
    r"""
    Choose steady solver depending on linear and flat flags
    """
    #[
    match (is_linear, is_flat, ):
        case (False, False, ):
            return _ft.partial(
                _steady_nonlinear,
                evaluator_class=_evaluators.NonflatSteadyEvaluator,
            )
        case (False, True, ):
            return _ft.partial(
                _steady_nonlinear,
                evaluator_class=_evaluators.FlatSteadyEvaluator,
            )
        case (True, False, ):
            return _ft.partial(
                _steady_linear,
                algorithm=_fs.solve_steady_linear_nonflat,
            )
        case (True, True, ):
            return _ft.partial(
                _steady_linear,
                algorithm=_fs.solve_steady_linear_flat,
            )
    #]


def _calculate_steady_incidence_matrix(
    equations: Iterable[Equation],
    wrt_any_qids: Iterable[int],
) -> _np.ndarray:
    r"""
    """
    #[
    qid_to_column = { qid: column for column, qid in enumerate(wrt_any_qids) }
    def token_within_quantities(tok: Token, ) -> bool:
        return qid_to_column.get(tok.qid, None)
        #
    return _equations.calculate_incidence_matrix(
        equations,
        len(wrt_any_qids),
        token_within_quantities,
    )
    #]


def _resolve_split_into_blocks(
    split_into_blocks: bool | None,
    plan: SteadyPlan | None,
) -> bool:
    r"""
    """
    #[
    if split_into_blocks is not None:
        return split_into_blocks
    if plan is None:
        return True
    any_fix = (
        plan.any_in_register("fixed_level", )
        or plan.any_in_register("fixed_change", )
    )
    return not any_fix
    #]


def _update_variant_with_final_guess(variant, steady_evaluator, qid_to_kind, qid_to_name, ) -> None:
    """
    """
    #[
    final_guess = steady_evaluator.final_guess
    #
    levels, wrt_level_qids = steady_evaluator.extract_levels(final_guess, )
    variant.update_levels_from_array(levels, wrt_level_qids, )
    #
    changes, wrt_change_qids = steady_evaluator.extract_changes(final_guess, )
    #
    # Filter out non-loggable quantities, i.e. endogenized parameters (only loggable
    # quantities can have steady-state change)
    changes_to_update = []
    wrt_change_qids_to_update = []
    for change, qid in zip(changes, wrt_change_qids, ):
        if qid_to_kind[qid] not in QuantityKind.LOGGABLE_VARIABLE:
            continue
        changes_to_update.append(change)
        wrt_change_qids_to_update.append(qid)
    variant.update_changes_from_array(changes_to_update, wrt_change_qids_to_update, )
    #]


def _throw_block_error(human_block, custom_header: str, ) -> NoReturn:
    """
    """
    #[
    message = (
        f"Steady state calculations failed to converge in {custom_header}",
        *human_block.equations,
        "Solved for " + " ".join(human_block.quantities, ),
    )
    raise _wrongdoings.IrisPieError(message, )
    #]





"""
Mixin for enumeration classes that represent keywords
"""


#[

from __future__ import annotations

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import Self

#]


class KindMixin:
    """
    """
    #[

    @classmethod
    def from_keyword(
        klass,
        keyword: str,
        /,
    ) -> Self:
        """
        """
        return klass[
            keyword
            .replace("-", "_")
            .replace(" ", "_")
            .replace("!", "")
            .strip()
            .upper()
            .removesuffix("S")
        ]

    def to_keyword(self, /, ) -> str:
        return "!" + self.name.lower() + "s"

    def _serialize_to_portable(self, /, ) -> str:
        return self.name

    @classmethod
    def _deserialize_from_portable(klass, value: str, /, ) -> Self:
        return klass[value]

    @property
    def human(self, /, ) -> str:
        return self.name.replace("_", " ").title()

    #]




r"""
"""

#[

from __future__ import annotations

#]


_DEFAULT_TOLERANCE = {
    "eigenvalue": 1e-12,
    "equality": 1e-12,
}


class InlayForInvariant:
    r"""
    """
    #[

    def reset_tolerance(self, ) -> None:
        r"""
        """
        self.tolerance = dict(_DEFAULT_TOLERANCE, )


class Inlay:
    r"""
    """
    #[

    def reset_tolerance(self, ) -> None:
        r"""
        """
        self._invariant.reset_tolerance()

    def override_tolerance(self, *args, **kwargs, ) -> dict[str, float]:
        r"""
        """
        tolerance = dict(self._invariant.tolerance, )
        update_tolerance = (args[0] if args else {}) | kwargs
        #
        invalid_keys = set(update_tolerance.keys()) - set(tolerance.keys())
        if invalid_keys:
            raise ValueError(f"Invalid tolerance keys: {invalid_keys}.", )
        #
        for key, value in update_tolerance.items():
            if value is not None:
                tolerance[key] = float(value)
        self._invariant.tolerance = tolerance
        return tolerance

    def get_tolerance(
        self,
        key: str | None = None,
    ) -> float | dict[str, float]:
        r"""
        """
        return self._invariant.tolerance[key] if key else self._invariant.tolerance

    #]




r"""
Broadcast functionality for time series
"""


#[

from __future__ import annotations

import numpy as _np
from .. import wrongdoings as _wrongdoings

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import Self
    from .main import Series

#]


def mixin(klass: type, ) -> type:
    r"""
    Inlay the broadcast methods in the class
    """
    #[
    klass.broadcast_variants = broadcast_variants
    return klass
    #]


#-------------------------------------------------------------------------------
# Functions to be used as methods in Series class
#-------------------------------------------------------------------------------


def broadcast_variants(self, num_variants, ) -> None:
    """
    Broadcast variants to match the specified number of variants
    """
    if self.data.shape[1] == num_variants:
        return
    if self.data.shape[1] == 1:
        self.data = _np.repeat(self.data, num_variants, axis=1, )
        return
    raise _wrongdoings.IrisPieError("Cannot broadcast variants")


#-------------------------------------------------------------------------------
# Standalone functions for use across modules
#-------------------------------------------------------------------------------


def broadcast_variants_when_needed(
    self: Series,
    other: Series,
) -> None:
    r"""
    Broadcast variants between two Series objects if needed
    """
    #[
    if self.num_variants == other.num_variants:
        return
    if self.num_variants == 1:
        self.broadcast_variants(other.num_variants, )
        return
    if other.num_variants == 1:
        other.broadcast_variants(self.num_variants, )
        return
    raise _wrongdoings.IrisPieError("Cannot broadcast time series variants")
    #]


#-------------------------------------------------------------------------------

